// Copyright 2022 The IREE Authors
//
// Licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

//===--------------- SetEncoding.cpp -------------------------------------===//
// Sets the encoding for compute operations to allow execution of the
// operations in tiled layouts.
//===---------------------------------------------------------------------===//

#include "iree-dialects/Dialect/LinalgExt/IR/LinalgExtDialect.h"
#include "iree-dialects/Dialect/LinalgExt/IR/LinalgExtOps.h"
#include "iree-dialects/Dialect/LinalgExt/Utils/Utils.h"
#include "iree/compiler/Dialect/Flow/Transforms/PassDetail.h"
#include "iree/compiler/Dialect/Flow/Transforms/Passes.h"
#include "mlir/Dialect/Affine/IR/AffineOps.h"
#include "mlir/Dialect/Arith/IR/Arith.h"
#include "mlir/Dialect/Arith/Utils/Utils.h"
#include "mlir/Dialect/Linalg/IR/Linalg.h"
#include "mlir/Dialect/MemRef/Transforms/Transforms.h"
#include "mlir/Dialect/Tensor/IR/Tensor.h"
#include "mlir/Dialect/Tensor/Utils/Utils.h"
#include "mlir/IR/BuiltinAttributeInterfaces.h"
#include "mlir/IR/BuiltinAttributes.h"
#include "mlir/IR/BuiltinTypes.h"
#include "mlir/IR/OpDefinition.h"
#include "mlir/IR/PatternMatch.h"
#include "mlir/IR/Types.h"
#include "mlir/Transforms/GreedyPatternRewriteDriver.h"

#include "llvm/Support/Debug.h"

namespace mlir {
namespace iree_compiler {
namespace IREE {
namespace Flow {

//===---------------------------------------------------------------------===//
// Utility functions
//===---------------------------------------------------------------------===//

/// Pads `value` enough for any actual tile sizes that could result from
/// materialization of `encodingAttr`.
static Value pad(OpBuilder &builder, Location loc, Value source,
                 IREE::LinalgExt::EncodingAttr encodingAttr) {
  RankedTensorType sourceType = source.getType().cast<RankedTensorType>();
  Type elemType = sourceType.getElementType();
  size_t rank = sourceType.getRank();
  RankedTensorType tensorTypeWithEncoding =
      RankedTensorType::get(sourceType.getShape(), elemType, encodingAttr);
  SmallVector<OpFoldResult> lowPad(rank, builder.getIndexAttr(0));
  SmallVector<Type> resultTypes(rank, builder.getIndexType());

  ValueRange encodingPaddingSizes =
      builder
          .create<IREE::LinalgExt::UpperBoundTileSizeOp>(
              loc, resultTypes, TypeAttr::get(tensorTypeWithEncoding))
          .getResults();
  SmallVector<OpFoldResult> highPad(rank);
  AffineExpr tileExpr, shapeExpr;
  bindSymbols(builder.getContext(), tileExpr, shapeExpr);
  AffineExpr highPadExpr = shapeExpr.ceilDiv(tileExpr) * tileExpr - shapeExpr;
  for (size_t i = 0; i < rank; ++i) {
    highPad[i] = affine::makeComposedFoldedAffineApply(
        builder, loc, highPadExpr,
        getAsOpFoldResult({encodingPaddingSizes[i],
                           builder.create<tensor::DimOp>(loc, source, i)}));
  }

  Value zero = builder.create<arith::ConstantOp>(loc, elemType,
                                                 builder.getZeroAttr(elemType));
  return builder.create<tensor::PadOp>(loc, /*resultType=*/nullptr, source,
                                       lowPad, highPad, zero);
}

static Value setEncoding(OpBuilder &builder, Location loc, Value source,
                         IREE::LinalgExt::EncodingAttr encodingAttr) {
  auto sourceType = source.getType().cast<RankedTensorType>();
  auto resultType = RankedTensorType::get(
      sourceType.getShape(), sourceType.getElementType(), encodingAttr);
  return builder.create<IREE::LinalgExt::SetEncodingOp>(loc, resultType,
                                                        source);
};

static LinalgExt::EncodingAttr makeEncoding(OpBuilder &builder,
                                            LinalgExt::EncodingUser user,
                                            LinalgExt::EncodingRole role,
                                            Type originalType) {
  auto *context = builder.getContext();
  auto userAttr = LinalgExt::EncodingUserAttr::get(context, user);
  auto roleAttr = LinalgExt::EncodingRoleAttr::get(context, role);
  auto originalTypeAttr =
      originalType ? TypeAttr::get(originalType) : TypeAttr{};
  return LinalgExt::EncodingAttr::get(context, userAttr, roleAttr,
                                      originalTypeAttr);
}

static Value padAndSetEncoding(OpBuilder &builder, Location loc, Value source,
                               LinalgExt::EncodingUser user,
                               LinalgExt::EncodingRole role) {
  // No need to specify original_type in the encoding poadded to pad(), because
  // the operand there is the `source` tensor, so it will default to reading its
  // original shape.
  auto encodingForPad =
      makeEncoding(builder, user, role, /*originalType=*/Type{});
  Value padded = pad(builder, loc, source, encodingForPad);
  // For setEncoding() below, we potentially need to specify an encoding with an
  // explicit original_type, because the operand there is the padded tensor
  // returned by pad() above, but we want setEncoding to be aware of the
  // original source tensor shape, not the padded tensor shape. To limit IR
  // verbosity, we only specify the original original_type when it differs from
  // the tensor type that the encoding is applied to.
  auto encodingForSetEncoding = encodingForPad;
  if (padded.getType() != source.getType()) {
    encodingForSetEncoding =
        makeEncoding(builder, user, role, source.getType());
  }
  return setEncoding(builder, loc, padded, encodingForSetEncoding);
}

static Value unsetEncodingAndExtractSlice(OpBuilder &builder, Location loc,
                                          Value source,
                                          SmallVector<OpFoldResult> sizes) {
  auto sourceType = source.getType().cast<RankedTensorType>();
  auto unsetEncodingReturnType =
      RankedTensorType::get(sourceType.getShape(), sourceType.getElementType());
  auto unsetEncoding = builder
                           .create<IREE::LinalgExt::UnsetEncodingOp>(
                               loc, unsetEncodingReturnType, source)
                           .getResult();
  auto rank = sourceType.getRank();
  SmallVector<OpFoldResult> offsets(rank, builder.getIndexAttr(0));
  SmallVector<OpFoldResult> strides(rank, builder.getIndexAttr(1));
  return builder.create<tensor::ExtractSliceOp>(loc, unsetEncoding, offsets,
                                                sizes, strides);
}

namespace {

/// Rewrites the matmul op to work on tensors with encoding. Optionally
/// also pads the operands.
struct SetMatmulEncoding : public OpRewritePattern<linalg::MatmulOp> {
  SetMatmulEncoding(MLIRContext *context, PatternBenefit benefit = 1)
      : OpRewritePattern<linalg::MatmulOp>(context, benefit) {}

  LogicalResult matchAndRewrite(linalg::MatmulOp matmulOp,
                                PatternRewriter &rewriter) const override {
    if (!matmulOp.hasTensorSemantics())
      return failure();
    auto inputs = matmulOp.getDpsInputs();
    auto outputs = matmulOp.getDpsInits();
    auto hasEncoding = [](Value operand) -> bool {
      auto type = llvm::dyn_cast<RankedTensorType>(operand.getType());
      return type && type.getEncoding();
    };
    if (llvm::any_of(inputs, hasEncoding) ||
        llvm::any_of(outputs, hasEncoding)) {
      return failure();
    }

    Value origLhs = inputs[0];
    Value origRhs = inputs[1];
    Value origOut = outputs[0];

    auto getElemType = [](Value v) -> Type {
      if (auto tensorType = llvm::dyn_cast<RankedTensorType>(v.getType())) {
        return tensorType.getElementType();
      }
      return {};
    };
    Type lhsElemType = getElemType(origLhs);
    Type rhsElemType = getElemType(origRhs);
    Type outElemType = getElemType(origOut);

    if (!lhsElemType || !rhsElemType || !outElemType) {
      return failure();
    }

    LinalgExt::EncodingUser user;

    if (lhsElemType.isF32() && rhsElemType.isF32() && outElemType.isF32()) {
      user = LinalgExt::EncodingUser::MATMUL_F32F32F32;
    } else if (lhsElemType.isF16() && rhsElemType.isF16() &&
               outElemType.isF32()) {
      user = LinalgExt::EncodingUser::MATMUL_F16F16F32;
    } else if (lhsElemType.isF16() && rhsElemType.isF16() &&
               outElemType.isF16()) {
      user = LinalgExt::EncodingUser::MATMUL_F16F16F16;
    } else if (lhsElemType.isBF16() && rhsElemType.isBF16() &&
               outElemType.isF32()) {
      user = LinalgExt::EncodingUser::MATMUL_BF16BF16F32;
    } else if (lhsElemType.isBF16() && rhsElemType.isBF16() &&
               outElemType.isBF16()) {
      user = LinalgExt::EncodingUser::MATMUL_BF16BF16BF16;
    } else if (lhsElemType.isSignlessInteger(8) &&
               rhsElemType.isSignlessInteger(8) &&
               outElemType.isSignlessInteger(32)) {
      user = LinalgExt::EncodingUser::MATMUL_I8I8I32;
    } else {
      return rewriter.notifyMatchFailure(
          matmulOp,
          "unhandled combination of (lhs, rhs, result) element types");
    }

    Location loc = matmulOp.getLoc();

    Value encodedLhs = padAndSetEncoding(rewriter, loc, origLhs, user,
                                         LinalgExt::EncodingRole::LHS);
    Value encodedRhs = padAndSetEncoding(rewriter, loc, origRhs, user,
                                         LinalgExt::EncodingRole::RHS);
    Value encodedOut = padAndSetEncoding(rewriter, loc, origOut, user,
                                         LinalgExt::EncodingRole::RESULT);

    Value matmulTiled = rewriter
                            .create<linalg::MatmulOp>(
                                loc, encodedOut.getType(),
                                ValueRange{encodedLhs, encodedRhs}, encodedOut)
                            .getResult(0);

    // Sizes are computed by original output size.
    FailureOr<SmallVector<OpFoldResult>> origOutSizes =
        LinalgExt::getDims(rewriter, loc, origOut);
    if (failed(origOutSizes)) {
      return rewriter.notifyMatchFailure(matmulOp,
                                         "failed to get shape of result");
    }

    Value result = unsetEncodingAndExtractSlice(rewriter, loc, matmulTiled,
                                                origOutSizes.value());

    rewriter.replaceOp(matmulOp, result);
    return success();
  }
};

struct SetBatchMatmulEncoding : public OpRewritePattern<linalg::BatchMatmulOp> {
  SetBatchMatmulEncoding(MLIRContext *context, PatternBenefit benefit = 1)
      : OpRewritePattern<linalg::BatchMatmulOp>(context, benefit) {}

  LogicalResult matchAndRewrite(linalg::BatchMatmulOp matmulOp,
                                PatternRewriter &rewriter) const override {
    if (!matmulOp.hasTensorSemantics())
      return failure();
    auto inputs = matmulOp.getDpsInputs();
    auto outputs = matmulOp.getDpsInits();
    auto hasEncoding = [](Value operand) -> bool {
      auto type = llvm::dyn_cast<RankedTensorType>(operand.getType());
      return type && type.getEncoding();
    };
    if (llvm::any_of(inputs, hasEncoding) ||
        llvm::any_of(outputs, hasEncoding)) {
      return failure();
    }

    Value origLhs = inputs[0];
    Value origRhs = inputs[1];
    Value origOut = outputs[0];

    auto getElemType = [](Value v) -> Type {
      if (auto tensorType = llvm::dyn_cast<RankedTensorType>(v.getType())) {
        return tensorType.getElementType();
      }
      return {};
    };
    Type lhsElemType = getElemType(origLhs);
    Type rhsElemType = getElemType(origRhs);
    Type outElemType = getElemType(origOut);

    if (!lhsElemType || !rhsElemType || !outElemType) {
      return failure();
    }

    LinalgExt::EncodingUser user;

    if (lhsElemType.isF32() && rhsElemType.isF32() && outElemType.isF32()) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_F32F32F32;
    } else if (lhsElemType.isF16() && rhsElemType.isF16() &&
               outElemType.isF32()) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_F16F16F32;
    } else if (lhsElemType.isF16() && rhsElemType.isF16() &&
               outElemType.isF16()) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_F16F16F16;
    } else if (lhsElemType.isBF16() && rhsElemType.isBF16() &&
               outElemType.isF32()) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_BF16BF16F32;
    } else if (lhsElemType.isBF16() && rhsElemType.isBF16() &&
               outElemType.isBF16()) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_BF16BF16BF16;
    } else if (lhsElemType.isSignlessInteger(8) &&
               rhsElemType.isSignlessInteger(8) &&
               outElemType.isSignlessInteger(32)) {
      user = LinalgExt::EncodingUser::BATCH_MATMUL_I8I8I32;
    } else {
      return rewriter.notifyMatchFailure(
          matmulOp,
          "unhandled combination of (lhs, rhs, result) element types");
    }

    Location loc = matmulOp.getLoc();

    Value encodedLhs = padAndSetEncoding(rewriter, loc, origLhs, user,
                                         LinalgExt::EncodingRole::LHS);
    Value encodedRhs = padAndSetEncoding(rewriter, loc, origRhs, user,
                                         LinalgExt::EncodingRole::RHS);
    Value encodedOut = padAndSetEncoding(rewriter, loc, origOut, user,
                                         LinalgExt::EncodingRole::RESULT);

    Value matmulTiled = rewriter
                            .create<linalg::BatchMatmulOp>(
                                loc, encodedOut.getType(),
                                ValueRange{encodedLhs, encodedRhs}, encodedOut)
                            .getResult(0);

    // Sizes are computed by original output size.
    FailureOr<SmallVector<OpFoldResult>> origOutSizes =
        LinalgExt::getDims(rewriter, loc, origOut);
    if (failed(origOutSizes)) {
      return rewriter.notifyMatchFailure(matmulOp,
                                         "failed to get shape of result");
    }

    Value result = unsetEncodingAndExtractSlice(rewriter, loc, matmulTiled,
                                                origOutSizes.value());

    rewriter.replaceOp(matmulOp, result);
    return success();
  }
};

/// Pattern to fold a `linalg.fill` -> `iree_linalg_ext.set_encoding`
/// operation into a `linalg.fill` of the encoded type.
struct FoldFillWithSetEncoding
    : public OpRewritePattern<IREE::LinalgExt::SetEncodingOp> {
  using OpRewritePattern<IREE::LinalgExt::SetEncodingOp>::OpRewritePattern;

  LogicalResult matchAndRewrite(IREE::LinalgExt::SetEncodingOp encodingOp,
                                PatternRewriter &rewriter) const override {
    auto fillOp = encodingOp.getSource().getDefiningOp<linalg::FillOp>();
    if (!fillOp)
      return failure();

    // Create a new fill op, with outs being defined by a new `tensor.empty` op.
    RankedTensorType encodingType = encodingOp.getResultType();
    Location loc = fillOp.getLoc();
    SmallVector<OpFoldResult> dimValues =
        tensor::getMixedSizes(rewriter, loc, fillOp.getOutputs()[0]);
    auto newEmptyOp = rewriter.create<tensor::EmptyOp>(
        loc, dimValues, encodingType.getElementType(),
        encodingType.getEncoding());
    rewriter.replaceOpWithNewOp<linalg::FillOp>(encodingOp, fillOp.getInputs(),
                                                ValueRange{newEmptyOp});
    return success();
  }
};

struct SetEncodingPass : public SetEncodingBase<SetEncodingPass> {
  void getDependentDialects(DialectRegistry &registry) const override {
    registry.insert<IREE::LinalgExt::IREELinalgExtDialect>();
  }

  void runOnOperation() override;
};
} // namespace

void SetEncodingPass::runOnOperation() {
  MLIRContext *context = &getContext();
  {
    RewritePatternSet patterns(context);
    patterns.insert<SetBatchMatmulEncoding, SetMatmulEncoding>(context);
    linalg::FillOp::getCanonicalizationPatterns(patterns, context);
    patterns.insert<FoldFillWithSetEncoding>(context);
    memref::populateResolveRankedShapedTypeResultDimsPatterns(patterns);
    if (failed(applyPatternsAndFoldGreedily(getOperation(),
                                            std::move(patterns)))) {
      return signalPassFailure();
    }
  }
}

std::unique_ptr<Pass> createSetEncodingPass() {
  return std::make_unique<SetEncodingPass>();
}

} // namespace Flow
} // namespace IREE
} // namespace iree_compiler
} // namespace mlir
